@@ -10,6 +10,7 @@ import autonetkit.load.graphml as graphml
 import autonetkit.exception
 import networkx as nx
 import os
+import autonetkit.ank as ank_utils
 
 __all__ = ['build']
 
@@ -85,7 +86,9 @@ def build(input_graph_string, timestamp):
     #update_messaging(anm)
     #build_conn(anm)
     build_ip(anm)
-    build_ip6(anm)
+
+    #TODO: make ip6 toggled by graph attribute
+    #build_ip6(anm)
     
     igp = G_in.data.igp or "ospf" #TODO: make default template driven
 #TODO: make the global igp be set on each node - this way can also support different IGPs per router
@@ -117,7 +120,6 @@ def boundary_nodes(G, nodes):
     """ returns nodes at boundary of G
     TODO: check works for both directed and undirected graphs
     based on edge_boundary from networkx """
-    import autonetkit.ank as ank_utils
     graph = ank_utils.unwrap_graph(G)
     nodes = list(nodes)
     nbunch = list(ank_utils.unwrap_nodes(nodes))
@@ -201,13 +203,12 @@ def build_bgp(anm):
             if not node.ibgp_l3_cluster or node.ibgp_l3_cluster == "None":
                 node.ibgp_l3_cluster = node.asn # ibgp_l3_cluster defaults to ASN
 
-        for asn, devices in G_phy.groupby("asn").items():
-            as_graph = G_phy.subgraph(devices)
-            routers = list(G_bgp.node(n) for n in as_graph if n.is_router)
+        for asn, devices in ank_utils.groupby("asn", G_bgp): # group by nodes in phy graph
+            routers = list(G_bgp.node(n) for n in devices if n.is_router) # list of nodes from bgp graph
 #TODO: catch integer cast exception
             ibgp_levels = set(int(r.ibgp_level) for r in routers)
             max_level = max(ibgp_levels)
-            all_pairs = [ (s, t) for s in routers for t in routers if s != t]
+            all_pairs = [ (s, t) for s in routers for t in routers if s != t] # all possible edge src/dst pairs
             if max_level == 3:
 
                 l1_l2_up_links = [ (s, t) for (s, t) in all_pairs if s.ibgp_level == 1 and t.ibgp_level == 2
@@ -221,8 +222,7 @@ def build_bgp(anm):
                 G_bgp.add_edges_from(l2_peer_links, type = 'ibgp', direction = 'over')
 
                 l2_l3_up_links = [ (s, t) for (s, t) in all_pairs if s.ibgp_level == 2 and t.ibgp_level == 3
-                        if s.ibgp_l2_cluster == t.ibgp_l2_cluster
-                        and s.ibgp_l3_cluster == t.ibgp_l3_cluster]
+                        if s.ibgp_l3_cluster == t.ibgp_l3_cluster]
                 l2_l3_down_links = [ (t, s) for (s, t) in l2_l3_up_links] # the reverse
                 G_bgp.add_edges_from(l2_l3_up_links, type = 'ibgp', direction = 'up')
                 G_bgp.add_edges_from(l2_l3_down_links, type = 'ibgp', direction = 'down')
@@ -231,22 +231,20 @@ def build_bgp(anm):
                 G_bgp.add_edges_from(l3_peer_links, type = 'ibgp', direction = 'over')
 
 # also check for any clusters which only contain l1 and l3 links
-                l3_clusters = set(r.ibgp_l3_cluster for r in routers)
-                for l3_cluster in l3_clusters:
-                    l3_cluster_devices = [r for r in routers if r.ibgp_l3_cluster == l3_cluster]
-                    l2_clusters = set(r.ibgp_l2_cluster for r in l3_cluster_devices)
-
-                    for l2_cluster in l2_clusters:
-                        l2_cluster_devices = [r for r in l3_cluster_devices if r.ibgp_l2_cluster == l2_cluster]
+                for l3_cluster, l3_cluster_devices in ank_utils.groupby("ibgp_l3_cluster", routers):
+                    for l2_cluster, l2_cluster_devices in ank_utils.groupby("ibgp_l2_cluster", l3_cluster_devices):
+                        l2_cluster_devices = list(l2_cluster_devices)
 
                         if any(r.ibgp_level == 2 for r in l2_cluster_devices):
-                            log.debug("L2 cluster %s in L3 cluster %s in ASN %s has l2 devices, not adding extra links" % (l2_cluster, l3_cluster, asn))
+                            log.debug("Cluster (%s, %s, %s) has l2 devices, not adding extra links" % (asn, l3_cluster, l2_cluster))
                         else:
                             l1_routers = [r for r in l2_cluster_devices if r.ibgp_level == 1]
                             l3_routers = [r for r in l2_cluster_devices if r.ibgp_level == 3]
-                            log.debug("L2 cluster %s in L3 cluster %s in ASN %s has no level 2 iBGP routers."
+                            if not(len(l1_routers) and len(l3_routers)):
+                                break # no routers to connect
+                            log.debug("Cluster (%s, %s, %s) has no level 2 iBGP routers. "
                                     "Connecting level 1 routers (%s) to level 3 routers (%s)"
-                                    % (l2_cluster, l3_cluster, asn, l1_routers, l3_routers))
+                                    % (asn, l3_cluster, l2_cluster, l1_routers, l3_routers))
 
                             l1_l3_up_links = [ (s, t) for s in l1_routers for t in l3_routers]
                             l3_l1_down_links = [ (t, s) for (s, t) in l1_l3_up_links] # the reverse
@@ -480,7 +478,6 @@ def build_ospf(anm):
                             (router.area, router, default_area))
                     router.area = default_area
 
-
     for router in G_ospf:
 # and set area on interface
         for edge in router.edges():
